// autogenerated by syzkaller (http://github.com/google/syzkaller)

#define _GNU_SOURCE

#include <arpa/inet.h>
#include <endian.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/futex.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_tun.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/if_arp.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/uio.h>
#include <unistd.h>

__attribute__((noreturn)) static void doexit(int status)
{
  volatile unsigned i;
  syscall(__NR_exit_group, status);
  for (i = 0;; i++) {
  }
}

#include <setjmp.h>
#include <signal.h>
#include <stdint.h>
#include <string.h>
#include <string.h>

const int kFailStatus = 67;
const int kRetryStatus = 69;

static void fail(const char* msg, ...)
{
  int e = errno;
  va_list args;
  va_start(args, msg);
  vfprintf(stderr, msg, args);
  va_end(args);
  fprintf(stderr, " (errno %d)\n", e);
  doexit((e == ENOMEM || e == EAGAIN) ? kRetryStatus : kFailStatus);
}

static __thread int skip_segv;
static __thread jmp_buf segv_env;

static void segv_handler(int sig, siginfo_t* info, void* uctx)
{
  uintptr_t addr = (uintptr_t)info->si_addr;
  const uintptr_t prog_start = 1 << 20;
  const uintptr_t prog_end = 100 << 20;
  if (__atomic_load_n(&skip_segv, __ATOMIC_RELAXED) &&
      (addr < prog_start || addr > prog_end)) {
    _longjmp(segv_env, 1);
  }
  doexit(sig);
}

static void install_segv_handler()
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_IGN;
  syscall(SYS_rt_sigaction, 0x20, &sa, NULL, 8);
  syscall(SYS_rt_sigaction, 0x21, &sa, NULL, 8);

  memset(&sa, 0, sizeof(sa));
  sa.sa_sigaction = segv_handler;
  sa.sa_flags = SA_NODEFER | SA_SIGINFO;
  sigaction(SIGSEGV, &sa, NULL);
  sigaction(SIGBUS, &sa, NULL);
}

#define NONFAILING(...)                                                \
  {                                                                    \
    __atomic_fetch_add(&skip_segv, 1, __ATOMIC_SEQ_CST);               \
    if (_setjmp(segv_env) == 0) {                                      \
      __VA_ARGS__;                                                     \
    }                                                                  \
    __atomic_fetch_sub(&skip_segv, 1, __ATOMIC_SEQ_CST);               \
  }

static void vsnprintf_check(char* str, size_t size, const char* format,
                            va_list args)
{
  int rv;

  rv = vsnprintf(str, size, format, args);
  if (rv < 0)
    fail("tun: snprintf failed");
  if ((size_t)rv >= size)
    fail("tun: string '%s...' doesn't fit into buffer", str);
}

static void snprintf_check(char* str, size_t size, const char* format,
                           ...)
{
  va_list args;

  va_start(args, format);
  vsnprintf_check(str, size, format, args);
  va_end(args);
}

#define COMMAND_MAX_LEN 128
#define PATH_PREFIX                                                    \
  "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin "
#define PATH_PREFIX_LEN (sizeof(PATH_PREFIX) - 1)

static void execute_command(const char* format, ...)
{
  va_list args;
  char command[PATH_PREFIX_LEN + COMMAND_MAX_LEN];
  int rv;

  va_start(args, format);
  memcpy(command, PATH_PREFIX, PATH_PREFIX_LEN);
  vsnprintf_check(command + PATH_PREFIX_LEN, COMMAND_MAX_LEN, format,
                  args);
  rv = system(command);
  if (rv != 0)
    fail("tun: command \"%s\" failed with code %d", &command[0], rv);

  va_end(args);
}

static int tunfd = -1;
static int tun_frags_enabled;

#define SYZ_TUN_MAX_PACKET_SIZE 1000

#define MAX_PIDS 32
#define ADDR_MAX_LEN 32

#define LOCAL_MAC "aa:aa:aa:aa:aa:%02hx"
#define REMOTE_MAC "bb:bb:bb:bb:bb:%02hx"

#define LOCAL_IPV4 "172.20.%d.170"
#define REMOTE_IPV4 "172.20.%d.187"

#define LOCAL_IPV6 "fe80::%02hxaa"
#define REMOTE_IPV6 "fe80::%02hxbb"

#define IFF_NAPI 0x0010
#define IFF_NAPI_FRAGS 0x0020

static void initialize_tun(uint64_t pid)
{
  if (pid >= MAX_PIDS)
    fail("tun: no more than %d executors", MAX_PIDS);
  int id = pid;

  tunfd = open("/dev/net/tun", O_RDWR | O_NONBLOCK);
  if (tunfd == -1) {
    printf(
        "tun: can't open /dev/net/tun: please enable CONFIG_TUN=y\n");
    printf("otherwise fuzzing or reproducing might not work as "
           "intended\n");
    return;
  }

  char iface[IFNAMSIZ];
  snprintf_check(iface, sizeof(iface), "syz%d", id);

  struct ifreq ifr;
  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, iface, IFNAMSIZ);
  ifr.ifr_flags = IFF_TAP | IFF_NO_PI | IFF_NAPI | IFF_NAPI_FRAGS;
  if (ioctl(tunfd, TUNSETIFF, (void*)&ifr) < 0) {
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    if (ioctl(tunfd, TUNSETIFF, (void*)&ifr) < 0)
      fail("tun: ioctl(TUNSETIFF) failed");
  }
  if (ioctl(tunfd, TUNGETIFF, (void*)&ifr) < 0)
    fail("tun: ioctl(TUNGETIFF) failed");
  tun_frags_enabled = (ifr.ifr_flags & IFF_NAPI_FRAGS) != 0;

  char local_mac[ADDR_MAX_LEN];
  snprintf_check(local_mac, sizeof(local_mac), LOCAL_MAC, id);
  char remote_mac[ADDR_MAX_LEN];
  snprintf_check(remote_mac, sizeof(remote_mac), REMOTE_MAC, id);

  char local_ipv4[ADDR_MAX_LEN];
  snprintf_check(local_ipv4, sizeof(local_ipv4), LOCAL_IPV4, id);
  char remote_ipv4[ADDR_MAX_LEN];
  snprintf_check(remote_ipv4, sizeof(remote_ipv4), REMOTE_IPV4, id);

  char local_ipv6[ADDR_MAX_LEN];
  snprintf_check(local_ipv6, sizeof(local_ipv6), LOCAL_IPV6, id);
  char remote_ipv6[ADDR_MAX_LEN];
  snprintf_check(remote_ipv6, sizeof(remote_ipv6), REMOTE_IPV6, id);

  execute_command("sysctl -w net.ipv6.conf.%s.accept_dad=0", iface);

  execute_command("sysctl -w net.ipv6.conf.%s.router_solicitations=0",
                  iface);

  execute_command("ip link set dev %s address %s", iface, local_mac);
  execute_command("ip addr add %s/24 dev %s", local_ipv4, iface);
  execute_command("ip -6 addr add %s/120 dev %s", local_ipv6, iface);
  execute_command("ip neigh add %s lladdr %s dev %s nud permanent",
                  remote_ipv4, remote_mac, iface);
  execute_command("ip -6 neigh add %s lladdr %s dev %s nud permanent",
                  remote_ipv6, remote_mac, iface);
  execute_command("ip link set dev %s up", iface);
}

static void setup_tun(uint64_t pid, bool enable_tun)
{
  if (enable_tun)
    initialize_tun(pid);
}

static void test();

void loop()
{
  while (1) {
    test();
  }
}

struct thread_t {
  int created, running, call;
  pthread_t th;
};

static struct thread_t threads[16];
static void execute_call(int call);
static int running;

static void* thr(void* arg)
{
  struct thread_t* th = (struct thread_t*)arg;
  for (;;) {
    while (!__atomic_load_n(&th->running, __ATOMIC_ACQUIRE))
      syscall(SYS_futex, &th->running, FUTEX_WAIT, 0, 0);
    execute_call(th->call);
    __atomic_fetch_sub(&running, 1, __ATOMIC_RELAXED);
    __atomic_store_n(&th->running, 0, __ATOMIC_RELEASE);
    syscall(SYS_futex, &th->running, FUTEX_WAKE);
  }
  return 0;
}

static void execute(int num_calls)
{
  int call, thread;
  running = 0;
  for (call = 0; call < num_calls; call++) {
    for (thread = 0; thread < sizeof(threads) / sizeof(threads[0]);
         thread++) {
      struct thread_t* th = &threads[thread];
      if (!th->created) {
        th->created = 1;
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_attr_setstacksize(&attr, 128 << 10);
        pthread_create(&th->th, &attr, thr, th);
      }
      if (!__atomic_load_n(&th->running, __ATOMIC_ACQUIRE)) {
        th->call = call;
        __atomic_fetch_add(&running, 1, __ATOMIC_RELAXED);
        __atomic_store_n(&th->running, 1, __ATOMIC_RELEASE);
        syscall(SYS_futex, &th->running, FUTEX_WAKE);
        struct timespec ts;
        ts.tv_sec = 0;
        ts.tv_nsec = 20 * 1000 * 1000;
        syscall(SYS_futex, &th->running, FUTEX_WAIT, 1, &ts);
        if (running)
          usleep((call == num_calls - 1) ? 10000 : 1000);
        break;
      }
    }
  }
}

long r[3];
uint64_t procid;
void execute_call(int call)
{
  switch (call) {
  case 0:
    syscall(__NR_mmap, 0x20000000, 0xfff000, 0x3, 0x32, 0xffffffff,
            0x0);
    break;
  case 1:
    NONFAILING(*(uint32_t*)0x2000a000 = 0x5);
    NONFAILING(*(uint32_t*)0x2000a004 = 0x78);
    NONFAILING(*(uint8_t*)0x2000a008 = 0x0);
    NONFAILING(*(uint8_t*)0x2000a009 = 0x0);
    NONFAILING(*(uint8_t*)0x2000a00a = 0x0);
    NONFAILING(*(uint8_t*)0x2000a00b = 0x0);
    NONFAILING(*(uint32_t*)0x2000a00c = 0x0);
    NONFAILING(*(uint64_t*)0x2000a010 = 0x0);
    NONFAILING(*(uint64_t*)0x2000a018 = 0x0);
    NONFAILING(*(uint64_t*)0x2000a020 = 0x0);
    NONFAILING(*(uint8_t*)0x2000a028 = 0x68);
    NONFAILING(*(uint8_t*)0x2000a029 = 0x0);
    NONFAILING(*(uint8_t*)0x2000a02a = 0x0);
    NONFAILING(*(uint8_t*)0x2000a02b = 0x0);
    NONFAILING(*(uint32_t*)0x2000a02c = 0x0);
    NONFAILING(*(uint32_t*)0x2000a030 = 0x0);
    NONFAILING(*(uint32_t*)0x2000a034 = 0x3);
    NONFAILING(*(uint64_t*)0x2000a038 = 0x20000000);
    NONFAILING(*(uint64_t*)0x2000a040 = 0x8);
    NONFAILING(*(uint64_t*)0x2000a048 = 0x0);
    NONFAILING(*(uint64_t*)0x2000a050 = 0x0);
    NONFAILING(*(uint64_t*)0x2000a058 = 0x0);
    NONFAILING(*(uint32_t*)0x2000a060 = 0x0);
    NONFAILING(*(uint64_t*)0x2000a068 = 0x0);
    NONFAILING(*(uint32_t*)0x2000a070 = 0x0);
    NONFAILING(*(uint16_t*)0x2000a074 = 0x0);
    NONFAILING(*(uint16_t*)0x2000a076 = 0x0);
    syscall(__NR_perf_event_open, 0x2000a000, 0x0, 0x0, 0xffffffff,
            0x0);
    break;
  case 2:
    syscall(__NR_time, 0x20000000);
    break;
  case 3:
    NONFAILING(*(uint32_t*)0x2033e000 = 0x4);
    syscall(__NR_getsockopt, 0xffffffff, 0x100, 0x1, 0x20082ffc,
            0x2033e000);
    break;
  case 4:
    NONFAILING(memcpy((void*)0x2052dff7, "/dev/kvm", 9));
    r[0] =
        syscall(__NR_openat, 0xffffffffffffff9c, 0x2052dff7, 0x0, 0x0);
    break;
  case 5:
    r[1] = syscall(__NR_ioctl, r[0], 0xae01, 0x0);
    break;
  case 6:
    r[2] = syscall(__NR_ioctl, r[1], 0xae41, 0x0);
    break;
  case 7:
    NONFAILING(*(uint64_t*)0x209a2f80 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2f88 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2f90 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2f98 = 0x2006);
    NONFAILING(*(uint64_t*)0x209a2fa0 = 0xf9ca);
    NONFAILING(*(uint64_t*)0x209a2fa8 = 0x44);
    NONFAILING(*(uint64_t*)0x209a2fb0 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fb8 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fc0 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fc8 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fd0 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fd8 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fe0 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2fe8 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2ff0 = 0x0);
    NONFAILING(*(uint64_t*)0x209a2ff8 = 0x0);
    syscall(__NR_ioctl, r[2], 0x4080aea2, 0x209a2f80);
    break;
  case 8:
    syscall(__NR_ioctl, r[2], 0xae80, 0x0);
    break;
  }
}

void test()
{
  memset(r, -1, sizeof(r));
  execute(9);
}

int main()
{
  int i;
  for (i = 0; i < 8; i++) {
    if (fork() == 0) {
      procid = i;
      install_segv_handler();
      setup_tun(i, true);
      loop();
      return 0;
    }
  }
  sleep(1000000);
  return 0;
}
